// file      : xsd/cxx/tree/tree-header.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/tree-header.hxx>
#include <cxx/tree/fundamental-header.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      // IDREF/IDREFS mapping.
      //
      struct IdRef : Traversal::Fundamental::IdRef,
                     Traversal::Fundamental::IdRefs,
                     protected virtual Context
      {
        IdRef (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_)
        {
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::IdRef& i)
        {
          // IdRef specialization is never named.
          //
          os << "typedef ::xsd::cxx::tree::idref< " <<
            type_name (i.argumented ().type ()) << ", " <<
            char_type << ", " <<
            xs_ns_name () << "::ncname > " << name << ";";
        }

        virtual Void
        traverse (SemanticGraph::Fundamental::IdRefs& i)
        {
          String xs_ns (xs_ns_name ());

          // IdRefs specialization is never named.
          //
          os << "typedef ::xsd::cxx::tree::idrefs< " <<
            char_type << ", " <<
            xs_ns << "::simple_type, " <<
            "::xsd::cxx::tree::idref< " <<
            type_name (i.argumented ().type ()) << ", " <<
            char_type << ", " <<
            xs_ns << "::ncname > > " << name << ";";
        }

      private:
        String
        type_name (SemanticGraph::Type& t)
        {
          // This type is always named.
          //
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (t);

          return o.str ();
        }

      private:
        String name;
      };


      // List mapping.
      //
      struct List : Traversal::List, protected virtual Context
      {
        List (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_),
              ex (c.esymbol.empty () ? c.esymbol : c.esymbol + L" ")
        {
        }

        virtual Void
        traverse (Type& l)
        {
          if (l.named ())
          {
            name = ename (l);

            // If renamed name is empty then we do not need to generate
            // anything for this type.
            //
            if (renamed_type (l, name) && !name)
              return;
          }


          os << "class " << ex << name <<
            ": public " << any_simple_type << ", " <<
            "public ::xsd::cxx::tree::list< " <<
            item_type_name (l) << ", " << char_type << " >"
             << "{"
             << "public:" << endl;

          // c-tor ()
          //
          os << name << " ();"
             << endl;

          // c-tor (istream&)
          //
          if (options.value<CLI::generate_extraction> ())
            os << "template <typename _istream>" << endl
               << name << " (::xsd::cxx::tree::istream< _istream >&," << endl
               << flags_type << " = 0," << endl
               << any_type << "* = 0);"
               << endl;

          if (!options.value<CLI::suppress_parsing> ())
          {
            // c-tor (xercesc::DOMElement)
            //
            os << name << " (const ::xercesc::DOMElement&," << endl
               << flags_type << " = 0," << endl
               << any_type << "* = 0);"
               << endl;

            // c-tor (xercesc::DOMAttr)
            //
            os << name << " (const ::xercesc::DOMAttr&," << endl
               << flags_type << " = 0," << endl
               << any_type << "* = 0);"
               << endl;

            // c-tor (std::basic_string const&, xercesc::DOMElement)
            //
            os << name << " (const " << string_type << "&," << endl
               << "const ::xercesc::DOMElement*," << endl
               << flags_type << " = 0," << endl
               << any_type << "* = 0);"
               << endl;
          }

          // copy c-tor ()
          //
          os << name << " (const " << name << "&," << endl
             << flags_type << " = 0," << endl
             << any_type << "* = 0);"
             << endl;

          os << "virtual " << name << "*" << endl
             << "_clone (" << flags_type << " = 0," << endl
             << any_type << "* = 0) const;"
             << endl;

          os << "};";
        }

      private:
        String
        item_type_name (Type& l)
        {
          // This type is always named.
          //
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (l.argumented ().type ());

          return o.str ();
        }

      private:
        String name;
        String ex;
      };


      // Union mapping.
      //
      struct Union : Traversal::Union, protected virtual Context
      {
        Union (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_),
              ex (c.esymbol.empty () ? c.esymbol : c.esymbol + L" ")
        {
        }

        virtual Void
        traverse (Type& u)
        {
          if (u.named ())
          {
            name = ename (u);

            // If renamed name is empty then we do not need to generate
            // anything for this type.
            //
            if (renamed_type (u, name) && !name)
              return;
          }

          os << "class " << ex << name <<
	    ": public " <<  xs_ns_name () << "::string"
             << "{"
             << "public:" << endl
             << endl;

          if (options.value<CLI::generate_default_ctor> ())
          {
            // c-tor ()
            //
            os << name << " ();"
               << endl;
          }

          // c-tor (string const&)
          //
          os << name << " (const " << string_type << "&);"
             << endl;

          // c-tor (istream&)
          //
          if (options.value<CLI::generate_extraction> ())
            os << "template <typename _istream>" << endl
               << name << " (::xsd::cxx::tree::istream< _istream >&," << endl
               << flags_type << " = 0," << endl
               << any_type << "* = 0);"
               << endl;

          if (!options.value<CLI::suppress_parsing> ())
          {
            // c-tor (xercesc::DOMElement)
            //
            os << name << " (const ::xercesc::DOMElement&," << endl
               << flags_type << " = 0," << endl
               << any_type << "* = 0);"
               << endl;

            // c-tor (xercesc::DOMAttr)
            //
            os << name << " (const ::xercesc::DOMAttr&," << endl
               << flags_type << " = 0," << endl
               << any_type << "* = 0);"
               << endl;

            // c-tor (std::basic_string const&, xercesc::DOMElement)
            //
            os << name << " (const " << string_type << "&," << endl
               << "const ::xercesc::DOMElement*," << endl
               << flags_type << " = 0," << endl
               << any_type << "* = 0);"
               << endl;
          }

          // copy c-tor ()
          //
          os << name << " (const " << name << "&," << endl
             << flags_type << " = 0," << endl
             << any_type << "* = 0);"
             << endl;

          os << "virtual " << name << "*" << endl
             << "_clone (" << flags_type << " = 0," << endl
             << any_type << "* = 0) const;"
             << endl;

          os << "};";
        }

      private:
        String name;
        String ex;
      };

      // Enum mapping.
      //
      struct Enumerator : Traversal::Enumerator, protected virtual Context
      {
        Enumerator (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          os << ename (e);
        }
      };


      struct Enumeration : Traversal::Enumeration, protected virtual Context
      {
        Enumeration (Context& c,
                     Traversal::NodeDispatcherBase& complex,
                     String const& name_ = L"")
            : Context (c),
              complex_ (complex),
              name (name_),
              ex (c.esymbol.empty () ? c.esymbol : c.esymbol + L" "),
              base_ (c),
              member_ (c),
              enumerator_ (c)
        {
          inherits_base_ >> base_;
          inherits_member_ >> member_;

          names_ >> enumerator_;
        }

        virtual Void
        traverse (Type& e)
        {
          if (e.named ())
          {
            name = ename (e);

            // If renamed name is empty then we do not need to generate
            // anything for this type.
            //
            if (renamed_type (e, name) && !name)
              return;
          }

          Boolean string_based (false);
          {
            IsStringBasedType t (string_based);
            t.dispatch (e);
          }

          String base_enum_name;
          Boolean enum_based (false);
          if (string_based)
          {
            SemanticGraph::Enumeration* be (0);
            IsEnumBasedType t (be);
            t.dispatch (e);

            if (be != 0)
            {
              enum_based = true;
              base_enum_name = ename (*be);
            }
          }

          UnsignedLong enum_count (0);

          for (Type::NamesIterator i (e.names_begin ()), end (e.names_end ());
               i != end; ++i)
            ++enum_count;

          os << "class " << ex << name << ": public ";

          // Enumeration always has a base.
          //
          inherits (e, inherits_base_);

          os << "{"
             << "public:" << endl;

          if (string_based)
          {
            if (enum_based)
            {
              os << "typedef _xsd_" << base_enum_name << " " <<
                "_xsd_" << name << ";";
            }
            else
            {
              os << "enum _xsd_" << name
                 << "{";

              enter_scope (name);

              names<Enumeration> (e, names_, 0, 0, 0, &Enumeration::comma);

              leave_scope ();

              os << "};";
            }
          }

          // default c-tor
          //
          if (options.value<CLI::generate_default_ctor> ())
          {
            os << name << " ();"
               << endl;
          }

          // c-tor (_xsd_name)
          //
          if (string_based)
          {
            os << name << " (_xsd_" << name << ");"
               << endl;
          }

          // c-tor (base)
          //
          os << name << " (const ";

          inherits (e, inherits_member_);

          os << "&);"
             << endl;

          // c-tor (istream&)
          //
          if (options.value<CLI::generate_extraction> ())
            os << "template <typename _istream>" << endl
               << name << " (::xsd::cxx::tree::istream< _istream >&," << endl
               << flags_type << " = 0," << endl
               << any_type << "* = 0);"
               << endl;

          if (!options.value<CLI::suppress_parsing> ())
          {
            // c-tor (xercesc::DOMElement)
            //
            os << name << " (const ::xercesc::DOMElement&," << endl
               << flags_type << " = 0," << endl
               << any_type << "* = 0);"
               << endl;

            // c-tor (xercesc::DOMAttr)
            //
            os << name << " (const ::xercesc::DOMAttr&," << endl
               << flags_type << " = 0," << endl
               << any_type << "* = 0);"
               << endl;

            // c-tor (std::basic_string const&, xercesc::DOMElement)
            //
            os << name << " (const " << string_type << "&," << endl
               << "const ::xercesc::DOMElement*," << endl
               << flags_type << " = 0," << endl
               << any_type << "* = 0);"
               << endl;
          }

          // copy c-tor
          //
          os << name << " (const " << name << "&," << endl
             << flags_type << " = 0," << endl
             << any_type << "* = 0);"
             << endl;

          os << "virtual " << name << "*" << endl
             << "_clone (" << flags_type << " = 0," << endl
             << any_type << "* = 0) const;"
             << endl;

          // operator= (_xsd_name)
          //
          if (string_based)
          {
            os << name << "&" << endl
               << "operator= (_xsd_" << name << ");"
               << endl;
          }

          // operator _xsd_name ()
          //
          // Lookup differences in various compilers and some really
          // twisted anonymous type hierarchies make generation of this
          // operator outside of the class a really hard task. So we are
          // going to make it "always inline".
          //
          if (string_based)
          {
            os << "virtual" << endl
               << "operator _xsd_" << name << " () const"
               << "{"
               << "return _xsd_" << name << "_convert ();"
               << "}";
          }

          //
          //
          if (string_based)
          {
            os << "protected:" << endl
               << "_xsd_" << name << endl
               << "_xsd_" << name << "_convert () const;"
               << endl
               << "public:" << endl;

            if (enum_based)
            {
              // We are going to reuse our base's literals.
              //
              os << "static const " << char_type << "* const* " <<
                "_xsd_" << name << "_literals_;";
            }
            else
            {
              os << "static const " << char_type << "* const " <<
                "_xsd_" << name << "_literals_[" << enum_count << "];";
            }

            os << "static const _xsd_" << name <<
              " _xsd_" << name << "_indexes_[" << enum_count << "];";
          }

          os << "};";
        }

        virtual Void
        comma (Type&)
        {
          os << "," << endl;
        }

      private:
        Traversal::NodeDispatcherBase& complex_;

        String name;
        String ex;

        Traversal::Inherits inherits_base_;
        BaseTypeName base_;

        Traversal::Inherits inherits_member_;
        MemberTypeName member_;

        Traversal::Names names_;
        Enumerator enumerator_;
      };


      //
      //
      struct MemberForward : Traversal::Member, protected virtual Context
      {
        MemberForward (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
	  if (skip (m)) return;

          String const& name (ename (m));

          if (name == L"type")
          {
            os << "// Workaround for g++ bug# 23206." << endl
               << "//" << endl
               << "public:" << endl
               << "struct " << name << ";"
               << endl;
          }
        }
      };


      // List of using declaration for all non-optional members.
      //
      struct UsingDecl : Traversal::Complex,
                         Traversal::Member,
                         protected virtual Context
      {
        UsingDecl (Context& c, String const& name)
            : Context (c), first_ (true), name_ (name)
        {
          inherits_ >> *this;
          names_ >> *this;
        }

        ~UsingDecl ()
        {
          if (!first_)
            os << "#endif" << endl
               << endl;
        }

        virtual Void
        traverse (SemanticGraph::Complex& c)
        {
          inherits (c, inherits_);

          enter_scope (ename (c));
          names (c, names_);
          leave_scope ();
        }

        virtual Void
        traverse (SemanticGraph::Member& m)
        {
          if (skip (m)) return;

          // Note that we are not going to include attributes with
          // default values here because they are going to be default-
          // initialized.
          //
          if (min (m) == 1 && max (m) == 1)
          {
            if (first_)
            {
              os << "#if defined(__EDG_VERSION__) || " <<
                "(defined(__HP_aCC) && __HP_aCC >= 60000)" << endl
                 << "public:" << endl;

              first_ = false;
            }

            os << "using " << name_ << "::base_::" << ename (m) << ";";
          }
        }

      private:
        Boolean first_;
        String name_;

      private:
        Traversal::Inherits inherits_;
        Traversal::Names names_;
      };


      //
      //
      struct MemberFunction: Traversal::Member, protected virtual Context
      {
        MemberFunction (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (skip (m)) return;

          String const& name (ename (m));

          Boolean fund (false);
          {
            IsFundamentalType t (fund);
            t.dispatch (m.type ());
          }


          if (max (m) != 1)
          {
            // sequence
            //
            String container (iscope_name (name, "container"));

            // container const&
            // name () const;
            //
            os << "const " << container << "&" << endl
               << name << " () const;"
               << endl;

            // container&
            // name ();
            //
            os << container << "&" << endl
               << name << " ();"
               << endl;

            // void
            // name (container const&);
            //
            os << "void" << endl
               << name << " (const " << container << "&);"
               << endl;
          }
          else if (min (m) == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
            String type (iscope_name (name, "type"));
            String container (iscope_name (name, "container"));

            // container const&
            // name () const;
            //
            os << "const " << container << "&" << endl
               << name << " () const;"
               << endl;

            // container&
            // name ();
            //
            os << container << "&" << endl
               << name << " ();"
               << endl;

            // void
            // name (type const&);
            //
            os << "void" << endl
               << name << " (const " << type << "&);"
               << endl;

            // void
            // name (container const&);
            //
            os << "void" << endl
               << name << " (const " << container << "&);"
               << endl;

            // void
            // name (auto_ptr<type>);
            //
            if (!fund)
              os << "void" << endl
                 << name << " (::std::auto_ptr< " << type << " >);"
                 << endl;
          }
          else
          {
            // one
            //
            String type (iscope_name (name, "type"));

            // type const&
            // name () const;
            //
            os << "const " << type << "&" << endl
               << name << " () const;"
               << endl;

            // type&
            // name ();
            //
            os << type << "&" << endl
               << name << " ();"
               << endl;

            // void
            // name (type const&);
            //
            os << "void" << endl
               << name << " (const " << type << "&);"
               << endl;

            // void
            // name (auto_ptr<type>);
            //
            if (!fund)
              os << "void" << endl
                 << name << " (::std::auto_ptr< " << type << " >);"
                 << endl;
          }
        }
      };

      template <typename Complex>
      struct Member : Traversal::Member, protected virtual Context
      {
        Member (Context& c)
            : Context (c),
              type_name_ (c),
              member_function_ (c)
        {
          belongs_ >> type_name_;
        }

        virtual Void
        traverse (Type& m)
        {
          if (skip (m)) return;

          String const& name (ename (m));

          os << "// " << m.name () << endl
             << "// " << endl;


          // For mapping.
          //
          os << "public:" << endl
             << "struct " << name
             << "{";

          SemanticGraph::Type& t (m.type ());

          if (t.named ())
          {
            os << "typedef ";

            belongs (m, belongs_);

            os << " " << unclash (name, "type") << ";";
          }
          else
          {
            if (t.context ().count ("seen"))
            {
              os << "typedef " <<
                t.context ().template get<String> ("anon-name") <<
                " " << unclash (name, "type") << ";";
            }
            else
            {
              os << "struct _xsd_" << name << "_"
                 << "{";

              Traversal::Belongs belongs_;

              IdRef id_ref (*this, name);
              List list (*this, name);
              Union union_ (*this, name);
              Complex complex (*this, name);
              Enumeration enumeration (*this, complex, name);

              belongs_ >> id_ref;
              belongs_ >> list;
              belongs_ >> union_;
              belongs_ >> complex;
              belongs_ >> enumeration;

              enter_scope (name);
              enter_scope (L"_xsd_" + name + L"_");

              t.context ().set ("seen", true);

              belongs (m, belongs_);

              t.context ().remove ("seen");

              leave_scope ();
              leave_scope ();

              os << "};";

              os << "typedef _xsd_" << name << "_::" <<
                name << " " << unclash (name, "type") << ";";
            }
          }

          os << "typedef ::xsd::cxx::tree::traits< " <<
            unclash (name, "type") << ", " << char_type << " > " <<
            unclash (name, "traits") << ";";

          if (max (m) != 1)
          {
            // sequence
            //
            os << "typedef ::xsd::cxx::tree::sequence< " <<
              unclash (name, "type") << " > " <<
              unclash (name, "container") << ";"

               << "typedef " << unclash (name, "container") <<
              "::iterator " << unclash (name, "iterator") << ";"

               << "typedef " << unclash (name, "container") <<
              "::const_iterator " << unclash (name, "const_iterator") << ";";

          }
          else if (min (m) == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
            os << "typedef ::xsd::cxx::tree::optional< " <<
              unclash (name, "type") << " > " <<
              unclash (name, "container") << ";";
          }
          else
          {
            // one
            //
          }


          // default_value
          //
          if (m.default_ ())
          {
            Boolean simple (true);

            if (m.is_a<SemanticGraph::Element> ())
            {
              IsSimpleType test (simple);
              test.dispatch (t);
            }

            if (simple)
            {
              os << endl
                 << "static const " << unclash (name, "type") << "&" << endl
                 << unclash (name, "default_value") << " ();"
                 << endl
                 << "private:" << endl
                 << "static const " << unclash (name, "type") << " " <<
                unclash (name, "default_value") << "_;";
            }
          }

          os << "};";

          member_function_.traverse (m);
        }

      private:
        MemberTypeName type_name_;
        Traversal::Belongs belongs_;

        MemberFunction member_function_;
      };



      struct DataMember: Traversal::Member, protected virtual Context
      {
        DataMember (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (skip (m)) return;

          String const& name (ename (m));

          if (max (m) != 1)
          {
            // sequence
            //
            os << "::xsd::cxx::tree::sequence< " <<
              iscope_name (name, "type") << " > " << "_xsd_" << name << "_;";
          }
          else if (min (m) == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            // optional
            //
            os << "::xsd::cxx::tree::optional< " <<
              iscope_name (name, "type") << " > " << "_xsd_" << name << "_;";
          }
          else
          {
            // one
            //
            os << "::xsd::cxx::tree::one< " <<
              iscope_name (name, "type") << " > " <<
              "_xsd_" << name << "_;";
          }
        }
      };


      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c, String const& name_ = L"")
            : Context (c),
              name (name_),
              e (c.esymbol.empty () ? c.esymbol : c.esymbol + L" "),
              base_name_ (c),
              member_name_ (c),
              member_forward_ (c),
              member_ (c),
              data_member_ (c),
              from_base_ctor_arg_ (c, false)
        {
          inherits_base_ >> base_name_;
          inherits_member_ >> member_name_;

          names_ >> member_;
          names_data_ >> data_member_;
          names_forward_ >> member_forward_;

          from_base_ctor_arg_names_ >> from_base_ctor_arg_;
        }

        virtual Void
        traverse (Type& c)
        {
          if (c.named ())
          {
            name = ename (c);

            // If renamed name is empty then we do not need to generate
            // anything for this type.
            //
            if (renamed_type (c, name) && !name)
              return;
          }

          Boolean has_members (has<Traversal::Member> (c));

          Boolean simple (true);
          {
            IsSimpleType t (simple);
            t.dispatch (c);
          }

          os << "class " << e << name << ": public ";

          if (c.inherits_p ())
            inherits (c, inherits_base_);
          else
            os << any_type;

          os << "{"
             << "public:" << endl
             << endl
             << "struct _xsd_" << name
             << "{"
             << "typedef ";

          if (c.inherits_p ())
            inherits (c, inherits_member_);
          else
            os << any_type;

          os << " base_;"
             << "};";

          enter_scope (name);
          c.context ().set ("anon-name", scope);

          // Forward declarations.
          //
          names (c, names_forward_);


          //
          //
          names (c, names_);

          c.context ().remove ("anon-name");
          leave_scope ();


          // Using declarations for all non-optional base members. They are
          // needed for EDG frontent-based compilers (Intel C++, aCC6, etc).
          //
          {
            UsingDecl complex (*this, L"_xsd_" + name);
            Traversal::Inherits inherits_complex (complex);

            inherits (c, inherits_complex);
          }

          os << "// Constructors." << endl
             << "//" << endl
             << "public:" << endl;

          // default c-tor
          //
          if (options.value<CLI::generate_default_ctor> ())
          {
            // c-tor (all-non-optional-members) will become default
            // c-tor if our inheritance hierarchy has no non-optional
            // members and no simple base.
            //
            Boolean generate (false);
            {
              GenerateDefaultCtor t (*this, generate);
              t.traverse (c);
            }

            if (generate)
            {
              os << name << " ();"
                 << endl;
            }
          }

          // c-tor (base, all-non-optional-members)
          //
          if (options.value<CLI::generate_from_base_ctor> ())
          {
            // c-tor (all-non-optional-members) will become
            // c-tor (base, all-non-optional-members) unless our
            // immediate base's hierarchy has some non-optional
            // members.
            //
            Boolean generate (false);
            {
              GenerateFromBaseCtor t (*this, generate);
              t.traverse (c);
            }

            if (generate)
            {
              String base_member (L"_xsd_" + name + L"::base_");

              os << name << " (const " << base_member << "&";

              enter_scope (name);
              names (c, from_base_ctor_arg_names_);
              leave_scope ();

              os << ");"
                 << endl;
            }
          }


          // c-tor (all-non-optional-members)
          //
          os << name << " (";

          {
            enter_scope (name);

            CtorArgs ctor_args (*this, L"_xsd_" + name, false);
            ctor_args.dispatch (c);

            leave_scope ();
          }

          os << ");"
             << endl;


          // c-tor (istream&)
          //
          if (options.value<CLI::generate_extraction> ())
            os << "template <typename _istream>" << endl
               << name << " (::xsd::cxx::tree::istream< _istream >&," << endl
               << flags_type << " = 0," << endl
               << any_type << "* = 0);"
               << endl;

          if (!options.value<CLI::suppress_parsing> ())
          {
            // c-tor (xercesc::DOMElement)
            //
            os << name << " (const ::xercesc::DOMElement&," << endl
               << flags_type << " = 0," << endl
               << any_type << "* = 0);"
               << endl;


            if (simple)
            {
              // c-tor (xercesc::DOMAttr)
              //
              os << name << " (const ::xercesc::DOMAttr&," << endl
                 << flags_type << " = 0," << endl
                 << any_type << "* = 0);"
                 << endl;

              // c-tor (std::basic_string const&, xercesc::DOMElement)
              //
              os << name << " (const " << string_type << "&," << endl
                 << "const ::xercesc::DOMElement*," << endl
                 << flags_type << " = 0," << endl
                 << any_type << "* = 0);"
                 << endl;
            }
          }

          // copy c-tor
          //
          os << name << " (const " << name << "&," << endl
             << flags_type << " = 0," << endl
             << any_type << "* = 0);"
             << endl;

          os << "virtual " << name << "*" << endl
             << "_clone (" << flags_type << " = 0," << endl
             << any_type << "* = 0) const;"
             << endl;


          // Data members and implementation functions.
          //
          if (has_members)
          {
            os << "// Implementation." << endl
               << "//" << endl
               << "private:" << endl;


            if (!options.value<CLI::suppress_parsing> ())
            {
              // parse (xercesc::DOMElement)
              //
              os << "void" << endl
                 << unclash (name, "parse") << " (" <<
                "const ::xercesc::DOMElement&, " << flags_type << ");"
                 << endl;
            }

            enter_scope (name);
            names (c, names_data_);
            leave_scope ();
          }

          os << "};";
        }

      private:
        String name;
        String e;

        Traversal::Inherits inherits_base_;
        BaseTypeName base_name_;

        Traversal::Inherits inherits_member_;
        MemberTypeName member_name_;

        Traversal::Names names_forward_;
        MemberForward member_forward_;

        Traversal::Names names_;
        Member<Complex> member_;

        Traversal::Names names_data_;
        DataMember data_member_;

        FromBaseCtorArg from_base_ctor_arg_;
        Traversal::Names from_base_ctor_arg_names_;
      };


      // Goes after anonymous types via global element. Note that we don't
      // need to go after attributes.
      //
      struct GlobalMember: Traversal::Element,
                           GlobalElementBase,
                           protected virtual Context
      {
        GlobalMember (Context& c)
            : GlobalElementBase (c), Context (c), type_name_ (c)
        {
          belongs_ >> type_name_;
        }

        virtual Void
        traverse (Type& m)
        {
          if (!generate_p (m))
            return;

          SemanticGraph::Type& t (m.type ());

          if (!t.named ())
          {
            String const& name (ename (m));

            os << "// " << m.name () << endl
               << "// " << endl;

            os << "class " << name
               << "{"
               << "struct _xsd_" << name << "_"
               << "{";

            Traversal::Belongs belongs_;

            IdRef id_ref (*this, name);
            List list (*this, name);
            Union union_ (*this, name);
            Complex complex (*this, name);
            Enumeration enumeration (*this, complex, name);

            belongs_ >> id_ref;
            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (name);
            enter_scope (L"_xsd_" + name + L"_");

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();

            os << "};"
               << "public:" << endl
               << "typedef _xsd_" << name << "_::" << name << " "
               << unclash (name, "type") << ";"
               << "};";
          }
        }

      private:
        MemberTypeName type_name_;
        Traversal::Belongs belongs_;
      };
    }

    Void
    generate_tree_header (Context& ctx,
                          String const& hxx_expr,
                          String const& ixx_expr,
                          String const& txx_expr)
    {
      String c (ctx.char_type);

      if (c == L"char")
      {
        ctx.os << "#ifndef XSD_USE_CHAR" << endl
               << "#define XSD_USE_CHAR" << endl
               << "#endif" << endl
               << endl;

        ctx.os << "#ifndef XSD_CXX_TREE_USE_CHAR" << endl
               << "#define XSD_CXX_TREE_USE_CHAR" << endl
               << "#endif" << endl
               << endl;
      }
      else if (c == L"wchar_t")
      {
        ctx.os << "#ifndef XSD_USE_WCHAR" << endl
               << "#define XSD_USE_WCHAR" << endl
               << "#endif" << endl
               << endl;

        ctx.os << "#ifndef XSD_CXX_TREE_USE_WCHAR" << endl
               << "#define XSD_CXX_TREE_USE_WCHAR" << endl
               << "#endif" << endl
               << endl;
      }

      if (ctx.options.value<CLI::generate_xml_schema> ())
      {
        ctx.os << "#include <xsd/cxx/tree/exceptions.hxx>" << endl
               << "#include <xsd/cxx/tree/elements.hxx>" << endl
               << "#include <xsd/cxx/tree/types.hxx>" << endl
               << endl;

        if (!ctx.options.value<CLI::suppress_parsing> () ||
            ctx.options.value<CLI::generate_serialization> ())
        {
          ctx.os << "#include <xsd/cxx/xml/error-handler.hxx>" << endl
                 << endl;
        }

        if (ctx.options.value<CLI::generate_serialization> ())
          ctx.os << "#include <xsd/cxx/xml/dom/namespace-infomap.hxx>" << endl
                 << endl;

        // Emit fundamental types.
        //
        {
          Traversal::Schema schema;
          Traversal::Names names;
          FundamentalNamespace ns (ctx);

          schema >> names >> ns;

          schema.dispatch (ctx.schema_root);
        }
      }
      else
      {
        Boolean inline_ (ctx.options.value<CLI::generate_inline> ());

        Boolean template_ (ctx.options.value<CLI::generate_insertion> () ||
                           ctx.options.value<CLI::generate_extraction> ());

        ctx.os << "#include <memory>    // std::auto_ptr" << endl
               << "#include <algorithm> // std::binary_search" << endl
               << endl;

        ctx.os << "#include <xsd/cxx/tree/exceptions.hxx>" << endl
               << "#include <xsd/cxx/tree/elements.hxx>" << endl
               << "#include <xsd/cxx/tree/containers.hxx>" << endl
               << "#include <xsd/cxx/tree/list.hxx>" << endl
               << endl;

        if (ctx.options.value<CLI::generate_extraction> ())
          ctx.os << "#include <xsd/cxx/tree/istream-fwd.hxx>" << endl
                 << endl;

        // Emit header includes.
        //
        {
          if (inline_ || template_)
          {
            ctx.os << "#ifndef XSD_DONT_INCLUDE_INLINE" << endl
                   << "#define XSD_DONT_INCLUDE_INLINE" << endl
                   << endl;
          }

          Traversal::Schema schema;
          Includes includes (ctx, hxx_expr);

          schema >> includes;

          schema.dispatch (ctx.schema_root);

          if (inline_ || template_)
          {
            ctx.os << "#undef XSD_DONT_INCLUDE_INLINE" << endl
                   << "#else" << endl
                   << endl;

            Traversal::Schema schema;
            Includes includes (ctx, hxx_expr);

            schema >> includes;

            schema.dispatch (ctx.schema_root);

            ctx.os << "#endif // XSD_DONT_INCLUDE_INLINE" << endl
                   << endl;
          }
        }


        {
          Traversal::Schema schema;

          Traversal::Sources sources;
          Traversal::Names names_ns, names;

          Namespace ns (ctx, true);

          List list (ctx);
          Union union_ (ctx);
          Complex complex (ctx);
          Enumeration enumeration (ctx, complex);
          GlobalMember member (ctx);

          schema >> sources >> schema;
          schema >> names_ns >> ns >> names;

          // Note that there is no IdRef since it is always anonymous.
          //
          names >> list;
          names >> union_;
          names >> complex;
          names >> enumeration;
          names >> member;

          schema.dispatch (ctx.schema_root);
        }

        // Emit inline includes.
        //
        if (inline_ || template_)
        {
          ctx.os << "#ifndef XSD_DONT_INCLUDE_INLINE" << endl
                 << endl;

          Traversal::Schema schema;
          Includes ixx_includes (ctx, ixx_expr);
          Includes txx_includes (ctx, txx_expr);

          if (inline_)
            schema >> ixx_includes;

          if (template_)
            schema >> txx_includes;

          schema.dispatch (ctx.schema_root);

          ctx.os << "#endif // XSD_DONT_INCLUDE_INLINE" << endl
                 << endl;
        }
      }
    }
  }
}
